import { describe, it } from 'node:test';
import assert from 'node:assert/strict';
import { formatType, parseTypeAndName, prependArg, toUpperSnakeCase } from '../common';

describe('Parsing Variable Declarations', () => {
    const tests: [string, string, string, string][] = [
        ['Simple', 'int x', 'int', 'x'],
        ['Pointer', 'int *y', 'int *', 'y'],
        ['Ugly Pointer', 'int* z', 'int*', 'z'],
        ['Double Pointer', 'int **a', 'int **', 'a'],
        ['Ugly Double Pointer', 'int* *b', 'int**', 'b'],
        ['Reference', 'int &c', 'int &', 'c'],
        ['Reference-To-Pointer', 'int *&d', 'int *&', 'd']
    ];
    for (const test of tests) {
        it(test[0], () => {
            const obj = parseTypeAndName(test[1]);
            assert.strictEqual(obj.type, test[2]);
            assert.strictEqual(obj.name, test[3]);
        });
    }
    it('Invalid', () => {
        assert.throws(() => {
            parseTypeAndName('abc');
        });
    });
});

describe('Upper-Snake-Case', () => {
    const tests: [string, string, string][] = [
        ['One Word', 'Hello', 'HELLO'],
        ['Two Words', 'HelloWorld', 'HELLO_WORLD'],
        ['Empty', '', ''],
        ['All Uppercase', 'HELLO', 'HELLO'],
        ['All Lowercase', 'hello', 'HELLO']
    ];
    for (const test of tests) {
        it(test[0], () => {
            assert.strictEqual(toUpperSnakeCase(test[1]), test[2]);
        });
    }
});

describe('Formatting Types For Concatenation', () => {
    const tests: [string, string, string][] = [
        ['Value', 'int', 'int '],
        ['Pointer', 'int *', 'int *'],
        ['Reference', 'int &', 'int &']
    ];
    for (const test of tests) {
        it(test[0], () => {
            assert.strictEqual(formatType(test[1]), test[2]);
        });
    }
});

describe('Prepending Arguments To Functions', () => {
    const tests: [string, string, string, string][] = [
        ['Empty', '()', 'int x', '(int x)'],
        ['Non-Empty', '(int x)', 'int y', '(int y, int x)']
    ];
    for (const test of tests) {
        it(test[0], () => {
            assert.strictEqual(prependArg(test[1], test[2]), test[3]);
        });
    }
});